Neko 1.99.9
A portable framework for high-order spectral element flow simulations
Loading...
Searching...
No Matches
opr_conv1.hip
Go to the documentation of this file.
1/*
2 Copyright (c) 2021-2026, The Neko Authors
3 All rights reserved.
4
5 Redistribution and use in source and binary forms, with or without
6 modification, are permitted provided that the following conditions
7 are met:
8
9 * Redistributions of source code must retain the above copyright
10 notice, this list of conditions and the following disclaimer.
11
12 * Redistributions in binary form must reproduce the above
13 copyright notice, this list of conditions and the following
14 disclaimer in the documentation and/or other materials provided
15 with the distribution.
16
17 * Neither the name of the authors nor the names of its
18 contributors may be used to endorse or promote products derived
19 from this software without specific prior written permission.
20
21 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
22 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
23 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
24 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
25 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
26 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
27 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
28 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
29 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
30 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
31 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
32 POSSIBILITY OF SUCH DAMAGE.
33*/
34
35#include <string.h>
36#include <stdlib.h>
37#include <stdio.h>
38#include <hip/hip_runtime.h>
40#include <device/hip/check.h>
41#include "conv1_kernel.h"
42#include "elem_block_tune.h"
43
44extern "C" {
45 #include <common/neko_log.h>
46}
47
48template < const int >
49int tune_conv1(void *du, void *u,
50 void *vx, void *vy, void *vz,
51 void *dx, void *dy, void *dz,
52 void *drdx, void *dsdx, void *dtdx,
53 void *drdy, void *dsdy, void *dtdy,
54 void *drdz, void *dsdz, void *dtdz,
55 void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel,
56 int *ch_sel, int *nwf_sel);
57
58extern "C" {
59
63 void hip_conv1(void *du, void *u,
64 void *vx, void *vy, void *vz,
65 void *dx, void *dy, void *dz,
66 void *drdx, void *dsdx, void *dtdx,
67 void *drdy, void *dsdy, void *dtdy,
68 void *drdz, void *dsdz, void *dtdz,
69 void *jacinv, int *nel, int *gdim, int *lx) {
70
71 static int autotune[17] = { 0 };
72 /* elements per block candidate chosen by the tuner */
73 static int autotune_eb[17] = { 0 };
74 /* chunk candidate chosen for the 1d variant */
75 static int autotune_ch[17] = { 0 };
76 /* wavefronts per block candidate chosen for the mfma variant */
77 static int autotune_nwf[17] = { 0 };
78
79 const dim3 nthrds_1d(1024, 1, 1);
80 const dim3 nthrds_kstep((*lx), (*lx), 1);
81 const dim3 nblcks((*nel), 1, 1);
82
83#define CASE_1D(LX, C) \
84 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
85 conv1_kernel_1d<real, LX, NEKO_CHUNKS(LX, C)> ), \
86 nblcks, NEKO_CHUNKS_NTHRDS(LX, C), 0, \
87 (hipStream_t) glb_cmd_queue, \
88 (real *) du, (real *) u, \
89 (real *) vx, (real *) vy, (real *) vz, \
90 (real *) dx, (real *) dy, (real *) dz, \
91 (real *) drdx, (real *) dsdx, (real *) dtdx, \
92 (real *) drdy, (real *) dsdy, (real *) dtdy, \
93 (real *) drdz, (real *) dsdz, (real *) dtdz, \
94 (real *) jacinv); \
95 HIP_CHECK(hipGetLastError());
96
97/* Runtime dispatch onto the tuned chunk candidate */
98#define CASE_1D_SEL(LX, SEL) \
99 switch (SEL) { \
100 case 0: CASE_1D(LX, 0); break; \
101 case 1: CASE_1D(LX, 1); break; \
102 case 2: CASE_1D(LX, 2); break; \
103 default: CASE_1D(LX, 3); break; \
104 }
105
106#define CASE_KSTEP(LX, C) \
107 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
108 conv1_kernel_kstep<real, LX, NEKO_EB(LX, C)> ), \
109 NEKO_EB_NBLCKS(*nel, LX, C), NEKO_EB_NTHRDS(LX, C), 0, \
110 (hipStream_t) glb_cmd_queue, \
111 (real *) du, (real *) u, \
112 (real *) vx, (real *) vy, (real *) vz, \
113 (real *) dx, (real *) dy, (real *) dz, \
114 (real *) drdx, (real *) dsdx, (real *) dtdx, \
115 (real *) drdy, (real *) dsdy, (real *) dtdy, \
116 (real *) drdz, (real *) dsdz, (real *) dtdz, \
117 (real *) jacinv, *nel); \
118 HIP_CHECK(hipGetLastError());
119
120/* Runtime dispatch onto the tuned candidate */
121#define CASE_KSTEP_SEL(LX, SEL) \
122 switch (SEL) { \
123 case 0: CASE_KSTEP(LX, 0); break; \
124 case 1: CASE_KSTEP(LX, 1); break; \
125 default: CASE_KSTEP(LX, 2); break; \
126 }
127
128#define CASE_MFMA(LX, C) \
129 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
130 conv1_kernel_mfma<real, LX, NEKO_MFMA_NWF(C), \
131 NEKO_MFMA_TILE(C)> ), \
132 NEKO_MFMA_NBLCKS(*nel, LX, C), NEKO_MFMA_NTHRDS(C), 0, \
133 (hipStream_t) glb_cmd_queue, \
134 (real *) du, (real *) u, \
135 (real *) vx, (real *) vy, (real *) vz, \
136 (real *) dx, (real *) dy, (real *) dz, \
137 (real *) drdx, (real *) dsdx, (real *) dtdx, \
138 (real *) drdy, (real *) dsdy, (real *) dtdy, \
139 (real *) drdz, (real *) dsdz, (real *) dtdz, \
140 (real *) jacinv, *nel); \
141 HIP_CHECK(hipGetLastError());
142
143/* Runtime dispatch onto the tuned candidate, which names both the wavefronts
144 per block and the matrix core tile, see NEKO_MFMA_TILE in mfma_kernel.h */
145#define CASE_MFMA_SEL(LX, SEL) \
146 switch (SEL) { \
147 case 0: CASE_MFMA(LX, 0); break; \
148 case 1: CASE_MFMA(LX, 1); break; \
149 case 2: CASE_MFMA(LX, 2); break; \
150 case 3: CASE_MFMA(LX, 3); break; \
151 case 4: CASE_MFMA(LX, 4); break; \
152 case 5: CASE_MFMA(LX, 5); break; \
153 case 6: CASE_MFMA(LX, 6); break; \
154 case 7: CASE_MFMA(LX, 7); break; \
155 case 8: CASE_MFMA(LX, 8); break; \
156 default: CASE_MFMA(LX, 9); break; \
157 }
158
159/* The switch above has to name every candidate, see the note in ax_helm.hip:
160 a candidate reaching the default arm launches one geometry's grid at
161 another geometry's kernel. */
162static_assert(NEKO_MFMA_CANDIDATES == 10,
163 "CASE_MFMA_SEL does not cover the candidate space");
164
165#define CASE(LX) \
166 case LX: \
167 if(autotune[LX] == 0 ) { \
168 autotune[LX]=tune_conv1<LX>(du, u, \
169 vx, vy, vz, \
170 dx, dy, dz, \
171 drdx, dsdx, dtdx, \
172 drdy, dsdy, dtdy, \
173 drdz, dsdz, dtdz, \
174 jacinv, nel, gdim, lx, &autotune_eb[LX], \
175 &autotune_ch[LX], \
176 &autotune_nwf[LX]); \
177 } else if (autotune[LX] == 1 ) { \
178 CASE_1D_SEL(LX, autotune_ch[LX]); \
179 } else if (autotune[LX] == 2 ) { \
180 CASE_KSTEP_SEL(LX, autotune_eb[LX]); \
181 } else if (autotune[LX] == 3 ) { \
182 CASE_MFMA_SEL(LX, autotune_nwf[LX]); \
183 } \
184 break
185
186#define CASE_LARGE(LX) \
187 case LX: \
188 CASE_KSTEP(LX, 0); \
189 break
190
191
192 if ((*lx) < 11) {
193 switch(*lx) {
194 CASE(2);
195 CASE(3);
196 CASE(4);
197 CASE(5);
198 CASE(6);
199 CASE(7);
200 CASE(8);
201 CASE(9);
202 CASE(10);
203 default:
204 {
205 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
206 exit(1);
207 }
208 }
209 }
210 else {
211 switch(*lx) {
212 CASE_LARGE(11);
213 CASE_LARGE(12);
214 CASE_LARGE(13);
215 CASE_LARGE(14);
216 CASE_LARGE(15);
217 CASE_LARGE(16);
218 default:
219 {
220 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
221 exit(1);
222 }
223 }
224 }
225 }
226}
227
228template < const int LX >
229int tune_conv1(void *du, void *u,
230 void *vx, void *vy, void *vz,
231 void *dx, void *dy, void *dz,
232 void *drdx, void *dsdx, void *dtdx,
233 void *drdy, void *dsdy, void *dtdy,
234 void *drdz, void *dsdz, void *dtdz,
235 void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel,
236 int *ch_sel, int *nwf_sel) {
239 int best1 = 0;
242 int best3 = 0;
243 const int rounds = neko_tune_rounds();
244 const int iters = neko_tune_iters();
245 /* Candidates of the kstep sweep, one -- the unblocked shape -- when the
246 elements per block sweep is off */
247 const int eb_cand = neko_eb_sweep() ? NEKO_EB_CANDIDATES : 1;
248 /* Geometry pinned by each formulation's own variable, -1 to sweep it */
249 const int ch_pin = neko_chunks_pin();
250 const int eb_pin = neko_eb_pin();
251 /* Candidates of the matrix core sweep, see neko_mfma_candidates() */
252 const int mfma_cand = neko_mfma_candidates();
253 /* Both matrix core dimensions, pinned by their own variables or -1 */
254 const int nwf_pin = neko_mfma_pin();
255 /* Formulation pinned by NEKO_AUTOTUNE, as the identifier this returns */
256 int strat = 0;
257 /* Hardware capability, which the explicit NEKO_AUTOTUNE=MFMA pin needs */
258 const bool mfma = mfma_lx_supported<LX>() && hip_have_mfma();
259 int best = 0;
260 int retval;
261
262 for (int c = 0; c < NEKO_EB_CANDIDATES; c++) {
264 }
265 for (int c = 0; c < NEKO_CHUNKS_CANDIDATES; c++) {
267 }
268 for (int c = 0; c < NEKO_MFMA_CANDIDATES; c++) {
270 }
271
272 const dim3 nthrds_1d(1024, 1, 1);
273 const dim3 nthrds_kstep((*lx), (*lx), 1);
274 const dim3 nblcks((*nel), 1, 1);
275 const hipStream_t stream = (hipStream_t) glb_cmd_queue;
276
277 char *env_value = NULL;
278 char neko_log_buf[80];
279
280 env_value=getenv("NEKO_AUTOTUNE");
281
282 sprintf(neko_log_buf, "Autotune conv1 (lx: %d)", *lx);
284
285 *eb_sel = 0;
286 *ch_sel = 0;
287 *nwf_sel = 0;
288
289 /*
290 * NEKO_AUTOTUNE names a formulation, and that is all it does: the sweep
291 * below is narrowed to that one kernel family, but its geometry -- the
292 * chunk size, the elements per block, the wavefronts per block -- is still
293 * measured candidate against candidate. A formulation this build or this
294 * device does not have is reported and ignored, leaving the full sweep.
295 */
296 if(env_value) {
297 if( !strcmp(env_value,"1D") ) {
298 strat = 1;
299 } else if( !strcmp(env_value,"KSTEP") ) {
300 strat = 2;
301 } else if( !strcmp(env_value,"MFMA") ) {
302 if (mfma) {
303 strat = 3;
304 } else {
305 sprintf(neko_log_buf, "MFMA strategy not available for this config");
307 }
308 } else {
309 sprintf(neko_log_buf, "Invalid value set for NEKO_AUTOTUNE");
311 }
312 }
313
314 /* Geometry of the pinned formulation, if its own variable fixes that too.
315 Both pinned leaves nothing to measure, so the kernel is launched once and
316 reported, which is what pinning has always done */
317 const int pin = (strat == 1) ? ch_pin : (strat == 2) ? eb_pin :
318 (strat == 3) ? nwf_pin : -1;
319
320 if (pin >= 0) {
321 switch (strat) {
322 case 1:
323 *ch_sel = pin;
325 sprintf(neko_log_buf, "Set by env : 1 (1D, %d chunk)",
327 break;
328 case 2:
329 *eb_sel = pin;
331 sprintf(neko_log_buf, "Set by env : 2 (KSTEP, %d elem/block)",
332 NEKO_EB_SEL(LX, pin));
333 break;
334 default:
335 *nwf_sel = pin;
337 sprintf(neko_log_buf, "Set by env : 3 (MFMA %s, %d wf, %d elem/blk)",
339 break;
340 }
343 return strat;
344 }
345
346 if (strat) {
347 sprintf(neko_log_buf, "Set by env : %d (%s)", strat, env_value);
349 }
350
351 /* Formulations the sweep considers, see NEKO_TUNE_FOR(). NEKO_MFMA_TUNE
352 keeps the matrix core variant out of an unpinned sweep, but naming it
353 explicitly still measures it */
354 const bool try_1d = (strat == 0 || strat == 1);
355 const bool try_kstep = (strat == 0 || strat == 2);
356 const bool try_mfma = mfma && ((strat == 3) ||
357 (strat == 0 && neko_mfma_sweep()));
358
361
362 /* Verify every candidate against the reference formulation on a synthetic
363 input before anything is timed, see neko_tune_verify_begin() in
364 elem_block_tune.h. A candidate that does not reproduce it is reported
365 and dropped from the comparison below. */
366 neko_tune_rej_t rej1 = {{0}}, rej2 = {{0}}, rej3 = {{0}};
367 {
369 real *vouts[] = { (real *) du };
370 void **vins[] = { &u, &vx, &vy, &vz };
371 if (neko_tune_verify_begin(&vf, vouts, 1, vins, 4,
372 (*nel) * LX * LX * LX, stream)) {
373 sprintf(neko_log_buf, "Verify : rel tol %.1e, synthetic input",
374 vf.tol);
376 CASE_1D_SEL(LX, 0);
377 neko_tune_verify_ref(&vf, stream);
380 }
383 }
386 }
387 neko_tune_verify_end(&vf, stream);
388 } else {
389 sprintf(neko_log_buf, "Verify : %s", neko_tune_verify_on() ?
390 "skipped, no memory for it" : "off (NEKO_TUNE_VERIFY=0)");
392 }
393 }
394
395 /* Warm every variant before timing anything: each specialisation has to be
396 resident and the clocks at steady state, or whichever is timed first is
397 measured on a colder part */
398 for (int i = 0; i < NEKO_TUNE_WARMUP; i++) {
400 CASE_1D_SEL(LX, c);
401 }
403 CASE_KSTEP_SEL(LX, c);
404 }
406 CASE_MFMA_SEL(LX, c);
407 }
408 }
409
410 /* Interleaved rounds, best time per variant */
411 for (int r = 0; r < rounds; r++) {
414 }
417 }
420 }
421 }
422
423 /* Candidates the verification pass rejected drop out here, after the
424 timing, so the sweep's own bookkeeping is untouched */
428
431
435 *eb_sel = best;
436 *ch_sel = best1;
437 *nwf_sel = best3;
438
439 if (time1[best1] < time2[best]) {
440 retval = 1;
441 } else {
442 retval = 2;
443 }
444
445 /* The mfma variant joins the comparison only where it exists, its
446 candidates are left at NEKO_TUNE_INIT otherwise */
447 if (time3[best3] < ((retval == 1) ? time1[best1] : time2[best])) {
448 retval = 3;
449 }
450
451 /* Leave the chosen kernel's output in place: the tuner stands in for a real
452 evaluation and the variants do not sum in the same order */
453 if (retval == 1) {
455 } else if (retval == 2) {
457 } else {
459 }
460
461 if (retval == 1) {
462 sprintf(neko_log_buf, "Chose : 1 (1D, %d chunk)",
464 } else if (retval == 2) {
465 sprintf(neko_log_buf, "Chose : 2 (KSTEP, %d elem/block)",
467 } else {
468 sprintf(neko_log_buf, "Chose : 3 (MFMA %s, %d wf, %d elem/blk)",
471 }
474 return retval;
475}
__global__ void ale_add_kinematics_kernel(const int n, T *__restrict__ wx, T *__restrict__ wy, T *__restrict__ wz, const T *__restrict__ x_ref, const T *__restrict__ y_ref, const T *__restrict__ z_ref, const T *__restrict__ phi, const T *__restrict__ x, const T *__restrict__ y, const T *__restrict__ z, const kinematics_params_t kin_params)
const int i
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dz
__global__ void const T *__restrict__ const T *__restrict__ dx
__global__ void const T *__restrict__ u
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ dy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ jacinv
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdy
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ vz
__global__ void const T *__restrict__ const T *__restrict__ vx
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ vy
#define NEKO_CHUNKS_CANDIDATES
Definition elem_block.h:125
#define NEKO_EB_CANDIDATES
Definition elem_block.h:63
#define NEKO_EB_SEL(LX, SEL)
Definition elem_block.h:108
#define NEKO_CHUNKS_SEL(LX, SEL)
Definition elem_block.h:147
static int neko_chunks_pin()
#define NEKO_TUNE_TIME(T, LAUNCH, LX, C, ITERS)
static int neko_tune_rounds()
static int neko_eb_pin()
#define NEKO_TUNE_LOG(LX, T1, T2)
#define NEKO_TUNE_INIT
#define NEKO_TUNE_BEST(T, BEST, N)
static int neko_tune_iters()
#define NEKO_TUNE_FOR(C, ON, PIN, N)
static int neko_eb_sweep()
#define NEKO_TUNE_WARMUP
double real
#define HIP_CHECK(err)
Definition check.h:8
static void neko_tune_verify_ref(neko_tune_verify_t *v, hipStream_t stream)
#define NEKO_TUNE_VERIFY_CAND(VF, REJ, LAUNCH, LX, C, NAME)
static void neko_tune_verify_end(neko_tune_verify_t *v, hipStream_t stream)
static int neko_tune_verify_on()
static int neko_tune_verify_begin(neko_tune_verify_t *v, real **outs, const int nout, void ***ins, const int nin, const int n, hipStream_t stream)
#define NEKO_TUNE_APPLY_REJ(T, REJ, N)
#define NEKO_TUNE_LOG_MFMA(LX, T3)
static const char * mfma_tile_name(const int c)
#define NEKO_MFMA_CANDIDATES
static int neko_mfma_pin()
static int neko_mfma_candidates()
static bool hip_have_mfma()
#define NEKO_MFMA_NWF(C)
#define NEKO_MFMA_EB(LX, C)
static int neko_mfma_sweep()
void log_error(char *msg)
void log_message(char *msg)
void log_end_section()
void log_section(char *msg)
#define CASE_KSTEP_SEL(LX, SEL)
#define CASE(LX)
int tune_conv1(void *du, void *u, void *vx, void *vy, void *vz, void *dx, void *dy, void *dz, void *drdx, void *dsdx, void *dtdx, void *drdy, void *dsdy, void *dtdy, void *drdz, void *dsdz, void *dtdz, void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel, int *ch_sel, int *nwf_sel)
#define CASE_1D_SEL(LX, SEL)
void hip_conv1(void *du, void *u, void *vx, void *vy, void *vz, void *dx, void *dy, void *dz, void *drdx, void *dsdx, void *dtdx, void *drdy, void *dsdy, void *dtdy, void *drdz, void *dsdz, void *dtdz, void *jacinv, int *nel, int *gdim, int *lx)
Definition opr_conv1.hip:63
#define CASE_MFMA_SEL(LX, SEL)
#define CASE_LARGE(LX)